<?php

namespace Nest\Model;

use Nest\Database\Exceptions\Query\MissingConditionValueException;
use Nest\Database\Exceptions\Query\NoPrimaryFieldException;
use Nest\Database\Exceptions\UnknownDatabaseException;
use Nest\Exceptions\InvalidTypeException;
use Nest\Model\Exceptions\MissingRequiredFieldException;
use Nest\Model\Exceptions\UndefinedRelationException;
use Nest\Types\Exceptions\IncompatibleTypeException;
use Nest\Model\Query\EntityQuery;
use function Nest\Utils\array_first;

/**
 * Class of a collection of entities, to perform operations on all of them.
 * @template Entity
 */
class Entities
{
	/**
	 * Entities.
	 * @var Entity[]
	 */
	protected array $entities;

	/**
	 * Create a collection of entities with an array of entities.
	 * @param Entity[] $entities
	 */
	public function __construct(array $entities)
	{
		$this->entities = $entities;
	}

	/**
	 * Get entities array.
	 * @return Entity[] The entities.
	 */
	public function get(): array
	{
		return $this->entities;
	}

	/**
	 * Set values to all the entities.
	 * @param string $name Name of the value to set.
	 * @param mixed $value Value to set.
	 * @return void
	 */
	public function __set(string $name, mixed $value): void
	{
		foreach ($this->entities as $entity)
			$entity->$name = $value;
	}

	/**
	 * Get all values with the given name in the entities.
	 * @param string $name Name of the value to get.
	 * @return array Array of values.
	 */
	public function __get(string $name): array
	{
		$result = [];
		foreach ($this->entities as $entity)
			$result[] = $entity->$name;
		return $result;
	}

	/**
	 * Call the given method on all entities.
	 * @param string $name Name of the method to call.
	 * @param array $arguments Arguments of the method.
	 * @return array Results of all calls.
	 */
	public function __call(string $name, array $arguments): array
	{
		$results = [];
		foreach ($this->entities as $entity)
			$results[] = $entity->$name(...$arguments);
		return $results;
	}

	/**
	 * Save current entities states.
	 * @return bool True if something was saved, false otherwise.
	 * @throws IncompatibleTypeException
	 * @throws MissingConditionValueException
	 * @throws MissingRequiredFieldException
	 * @throws NoPrimaryFieldException
	 * @throws InvalidTypeException
	 */
	public function save(): bool
	{
		$result = false;
		foreach ($this->entities as $entity)
			// Set result to true if needed.
			$result = $result || $entity->save();
		return $result;
	}

	/**
	 * Load given relation in the entities.
	 * @param array<string|int, string|callable(EntityQuery): void> $relations Relations to load.
	 * @return $this
	 * @throws UnknownDatabaseException
	 * @throws IncompatibleTypeException
	 * @throws InvalidTypeException
	 * @throws MissingConditionValueException
	 * @throws UndefinedRelationException
	 */
	public function load(array $relations): static
	{
		// Do nothing if there are no entities.
		if (empty($this->entities)) return $this;

		// Normalize relations array.
		$relations = EntityQuery::normalizeRelationsDefinition($relations);

		// Perform load for all the entities.
		array_first($this->entities)->query()->load($this->entities, $relations);

		return $this;
	}
}
