<?php

namespace Nest\Model;

use Carbon\Carbon;
use Nest\Application;
use Nest\Database\Database;
use Nest\Database\Query\QueryBuilder;
use Nest\Database\Query\Raw;
use Nest\Events\HasEvents;
use Nest\Database\Exceptions\Query\MissingConditionValueException;
use Nest\Database\Exceptions\Query\NoPrimaryFieldException;
use Nest\Database\Exceptions\UnknownDatabaseException;
use Nest\Exceptions\InvalidTypeException;
use Nest\Model\Exceptions\MissingRequiredFieldException;
use Nest\Model\Exceptions\UndefinedRelationException;
use Nest\Model\Exceptions\UnhandledPropertyTypeException;
use Nest\Types\Exceptions\IncompatibleTypeException;
use Nest\Model\Events\AfterDelete;
use Nest\Model\Events\AfterInsert;
use Nest\Model\Events\AfterLoad;
use Nest\Model\Events\AfterPropertiesInitialization;
use Nest\Model\Events\AfterSave;
use Nest\Model\Events\AfterUpdate;
use Nest\Model\Events\BeforeDelete;
use Nest\Model\Events\BeforeInsert;
use Nest\Model\Events\BeforeLoad;
use Nest\Model\Events\BeforePropertiesInitialization;
use Nest\Model\Events\BeforeSave;
use Nest\Model\Events\BeforeUpdate;
use Nest\Model\Events\EntityDefinitionEvent;
use Nest\Model\Query\EntityQuery;
use Nest\Types\SqlType;
use function Nest\Utils\array_first;
use function Nest\Utils\get_class_name_from_fullname;
use function Nest\Utils\str_camel_to_snake;
use function Nest\Utils\str_snake_pluralize;

/**
 * Class of an entity of the model.
 */
abstract class Entity
{
	use HasEvents;
	/**
	 * Set if the entity is new (should be inserted) or not.
	 * @var bool
	 */
	protected bool $isNew = true;

	/**
	 * Original properties values.
	 * @var array
	 */
	protected array $originals = [];

	/**
	 * Entity definition cache.
	 * @var EntityBlueprint
	 */
	private EntityBlueprint $_definition;

	/**
	 * Definition of the entity.
	 * @param EntityBlueprint $blueprint Entity blueprint to define.
	 * @return EntityBlueprint Defined entity blueprint.
	 */
	public abstract function definition(EntityBlueprint $blueprint): EntityBlueprint;

	/**
	 * Get the entity definition (from cache if already requested).
	 * @return EntityBlueprint Defined entity blueprint.
	 */
	private function _getDefinition(): EntityBlueprint
	{
		if (empty($this->_definition))
		{ // Get a new definition and save it in cached definition.
			$this->_definition = $this->definition(new EntityBlueprint(static::class));
			$this->getEventsManager()->fire(new EntityDefinitionEvent($this));
		}

		// Return cached definition.
		return $this->_definition;
	}

	/**
	 * Initialize entity.
	 */
	public function __construct()
	{
		// Initialize properties.
		$this->initializeProperties();
	}

	/**
	 * Initialize entity default properties.
	 * @return void
	 */
	protected function initializeProperties(): void
	{
		$this->getEventsManager()->fire(new BeforePropertiesInitialization($this));

		foreach ($this->_getDefinition()->getProperties() as $propertyName => $property)
		{ // For each property, set its default value if there is one.
			if ($property instanceof ReadableFieldBlueprint && $property->hasCurrentDateByDefault())
				$this->$propertyName = Carbon::now();
		}

		$this->getEventsManager()->fire(new AfterPropertiesInitialization($this));
	}


	/**
	 * Guess the entity database table name from its class name.
	 * Always use a plural snake_cased name.
	 * @return string The database table name of the entity, based on its class name.
	 */
	private function guessTableName(): string
	{
		return str_snake_pluralize(str_camel_to_snake(get_class_name_from_fullname(
			(new \ReflectionClass($this))->isAnonymous() ? get_parent_class(static::class) : static::class,
		)));
	}

	/**
	 * Get the entity database table name.
	 * @return string The database table name of the entity.
	 */
	public function getTableName(): string
	{
		return $this->_getDefinition()->getTable() ?? $this->guessTableName();
	}

	/**
	 * Check that the required fields are set.
	 * @return string[] Names of the missing required fields.
	 */
	public function getMissingRequiredFields(): array
	{
		// Get current entity definition.
		$entityBlueprint = $this->_getDefinition();
		// Get its required fields.
		$requiredFields = array_filter($entityBlueprint->getProperties(), fn ($field) => $field instanceof ReadableFieldBlueprint && $field->isRequired());

		// Initialize missing required fields list.
		$missingRequiredFields = [];

		foreach ($requiredFields as $requiredFieldName => $requiredField)
		{ // If any required field is not set, add it in the list.
			if (!isset($this->{$requiredFieldName}))
				// A required field is not set.
				$missingRequiredFields[] = $requiredFieldName;
		}

		return $missingRequiredFields; // All the missing required fields.
	}

	/**
	 * Check that all the required fields are set.
	 * Throw an exception if some are not set.
	 * @return void
	 * @throws MissingRequiredFieldException
	 */
	public function checkRequiredFields(): void
	{
		if (!empty($missingFields = $this->getMissingRequiredFields()))
			// Some required fields are missing, throwing an exception.
			throw new MissingRequiredFieldException($missingFields);
	}

	/**
	 * Get primary fields.
	 * @return string|string[] The primary field name, or array of fields when there are multiple primary fields.
	 */
	public function getPrimaryFields(): string|array
	{
		// Search primary fields.
		$primaryFields = array_filter($this->_getDefinition()->getProperties(), fn ($field) => $field instanceof ReadableFieldBlueprint && $field->isPrimary());

		// Return primary field(s).
		return count($primaryFields) == 1 ? array_key_first($primaryFields) : array_keys($primaryFields);
	}


	/**
	 * Get database of the entity.
	 * @return Database Database of the entity.
	 * @throws UnknownDatabaseException
	 */
	public function getDatabase(): Database
	{
		return Application::get()->databases()->db($this->_getDefinition()->getDatabase());
	}

	/**
	 * Create a new SQL query for the entity.
	 * @return QueryBuilder SQL query builder.
	 * @throws UnknownDatabaseException
	 */
	public function sql(): QueryBuilder
	{
		return $this->getDatabase()->query($this->getTableName());
	}

	/**
	 * Create a new entity query.
	 * @return EntityQuery<static> Entity query builder.
	 * @throws UnknownDatabaseException
	 */
	public function query(): EntityQuery
	{
		// Create the query.
		$query = new EntityQuery($this->getDatabase(), $this->getTableName(), $this);

		// Automatically fill "with" with eager loaded properties.
		$query->with(array_filter(array_map(fn ($property) => (
			$property instanceof ReadableEntityPropertyBlueprint && $property->doEagerLoad()
		), $this->_getDefinition()->getProperties())));

		return $query; // Return the created query.
	}

	/**
	 * Load given relations in the entity.
	 * @param array<string|int, string|callable(EntityQuery): void> $relations Relations to load.
	 * @return $this
	 * @throws UnknownDatabaseException
	 * @throws IncompatibleTypeException
	 * @throws InvalidTypeException
	 * @throws MissingConditionValueException
	 * @throws UndefinedRelationException
	 * @throws UnhandledPropertyTypeException
	 */
	public function load(array $relations): static
	{
		// Normalize relations array.
		$relations = EntityQuery::normalizeRelationsDefinition($relations);

		$this->getEventsManager()->fire(new BeforeLoad($this, $relations));

		// Perform load for the current entity.
		$this->query()->load([$this], $relations);

		$this->getEventsManager()->fire(new AfterLoad($this, $relations));

		return $this;
	}

	/**
	 * Create a new entity of the same type.
	 * @return static A new entity of the same type.
	 */
	public function new(): static
	{
		return new static();
	}

	/**
	 * Set an original property value.
	 * @param string $propertyName Property name.
	 * @param mixed $propertyValue Property value.
	 * @return void
	 */
	public function setOriginalProperty(string $propertyName, mixed $propertyValue): void
	{
		$this->$propertyName = $propertyValue;
		$this->originals[$propertyName] = $propertyValue;
	}

	/**
	 * Find changes made on the entity.
	 * @return array<string, mixed> New properties values.
	 */
	public function diff(): array
	{
		// Initialize diff array.
		$diff = [];

		foreach ($this->_getDefinition()->getProperties() as $propertyName => $propertyDefinition)
		{ // Add current property value to diff if it's different from original value.
			if (!($propertyDefinition instanceof FieldBlueprint || $propertyDefinition instanceof ArrayBlueprint))
				// Only keep normal fields.
				continue;

			if (isset($this->$propertyName) && (!isset($this->originals[$propertyName]) || $this->originals[$propertyName] != $this->$propertyName))
				// Current value is different from original one, adding it to the diff.
				$diff[$propertyName] = $this->$propertyName;
		}

		return $diff; // Return properties diff.
	}

	/**
	 * Save the current entity state.
	 * @return bool True if something was saved, false otherwise.
	 * @throws IncompatibleTypeException
	 * @throws MissingConditionValueException
	 * @throws MissingRequiredFieldException
	 * @throws NoPrimaryFieldException
	 * @throws UnhandledPropertyTypeException
	 * @throws UnknownDatabaseException
	 */
	public function save(): bool
	{
		$this->getEventsManager()->fire(new BeforeSave($this));

		// Check required fields to save.
		$this->checkRequiredFields();

		if ($this->isNew)
		{ // This is a new entity, inserting it, then reading the result.
			$this->fromSqlProperties($this->insert());
		}
		else
		{ // The entity already exists, updating it.

			// Get entity diff.
			$diff = $this->diff();

			// No diff, return false because there is nothing to update.
			if (empty($diff)) return false;

			// Get properties definition.
			$properties = $this->_getDefinition()->getProperties();

			foreach ($diff as $propertyName => $propertyValue)
			{ // For each property in diff, check if it's an array field.
				if ($properties[$propertyName] instanceof ArrayBlueprint)
				{ // If the property is an array field, save it then remove it from diff.
					(new ArrayInstance($properties[$propertyName]))->save($this, $propertyValue);
					unset($diff[$propertyName]);
				}
			}

			// Update every else property (if there are still some).
			if (!empty($diff))
				$this->fromSqlProperties($this->update($diff));
		}

		$this->getEventsManager()->fire(new AfterSave($this));

		return true; // Entity saved.
	}

	/**
	 * Delete the entity.
	 * @return bool
	 * @throws MissingConditionValueException
	 * @throws NoPrimaryFieldException
	 * @throws UnknownDatabaseException
	 */
	public function delete(): bool
	{
		$this->getEventsManager()->fire(new BeforeDelete($this));

		// Delete current entity.
		$this->sql()->newDelete()->whereKeyOf($this)->execute();

		// Set entity as new, in case of a save after deletion.
		$this->isNew = true;

		$this->getEventsManager()->fire(new AfterDelete($this));

		return true; // Entity is deleted.
	}

	/**
	 * Insert current entity values.
	 * @return object Raw inserted object.
	 * @throws IncompatibleTypeException
	 * @throws UnknownDatabaseException
	 */
	private function insert(): object
	{
		$this->getEventsManager()->fire(new BeforeInsert($this));

		// Return the first result (= the sole inserted row).
		$result = array_first(
			// Execute INSERT in RETURNING mode.
			$this->sql()->newInsert()->values(
				// Extract properties in an SQL row values array.
				$this->toSqlProperties()
			)->execute(true)
		);

		$this->getEventsManager()->fire(new AfterInsert($this));

		return $result;
	}

	/**
	 * Update given entity values.
	 * @param array<string, mixed> $values Values to update.
	 * @return object Raw updated object.
	 * @throws MissingConditionValueException
	 * @throws IncompatibleTypeException
	 * @throws NoPrimaryFieldException
	 * @throws UnknownDatabaseException
	 */
	private function update(array $values): object
	{
		$this->getEventsManager()->fire(new BeforeUpdate($this));

		// Return the first result (= the sole updated row).
		$result = array_first(
			// Execute UPDATE in RETURNING mode.
			$this->sql()->newUpdate()->set($this->valuesToSql($values))
				->whereKeyOf($this)
				->execute(true)
		);

		$this->getEventsManager()->fire(new AfterUpdate($this));

		return $result;
	}

	/**
	 * Get relation of the given name.
	 * @param string $relationName Name of the relation to get.
	 * @return EntityPropertyInstance The property instance.
	 * @throws UndefinedRelationException
	 */
	public function getRelation(string $relationName): EntityPropertyInstance
	{
		// Get defined properties for the entity.
		$properties = $this->_getDefinition()->getProperties();

		if (!empty($properties[$relationName]))
		{ // A property with the relation name exists, trying to get its relation definition.
			$definition = $properties[$relationName];
			if (is_a($definition, EntityPropertyBlueprint::class))
				// Initialize a property instance.
				return new EntityPropertyInstance($definition);
			else
			{ // Not an entity blueprint, throw an exception.
				throw new UndefinedRelationException(static::class, $relationName);
			}
		}
		else
		{ // No property with this name, throw an exception.
			throw new UndefinedRelationException(static::class, $relationName);
		}
	}

	/**
	 * Convert given values to SQL from matching properties types.
	 * @param array<string, mixed> $values Values to convert to SQL value.
	 * @return array<string, Raw|string|int|float|null> SQL values.
	 * @throws IncompatibleTypeException
	 */
	private function valuesToSql(array $values): array
	{
		// Get properties definition.
		$propertiesDefinition = $this->_getDefinition()->getProperties();

		foreach ($values as $propertyName => $value)
		{ // For each value, try to convert its data.

			// Keep the value (and convert it) if there is a matching property.
			if (!empty($propertiesDefinition[$propertyName]))
			{
				// Get property type.
				$propertyType = $propertiesDefinition[$propertyName]->getTypeInstance();

				if (!$propertyType instanceof SqlType)
					// Type is not an SQL type, throwing an exception.
					throw new IncompatibleTypeException(get_class($propertyType), SqlType::class);

				// Converting property value and put it in the SQL object.
				$values[$propertyName] = $propertyType->toSql($value);
			}
			else
				// No property matching the current value, unset it.
				unset($values[$propertyName]);
		}

		return $values; // Return converted values.
	}

	/**
	 * Import properties from a raw SQL object.
	 * @param object $sqlProperties Raw SQL object where to extract properties.
	 * @param string $prefix Optional prefix of properties in SQL.
	 * @return void
	 * @throws IncompatibleTypeException
	 * @throws UnhandledPropertyTypeException
	 */
	public function fromSqlProperties(object $sqlProperties, string $prefix = ""): void
	{
		$this->isNew = false; // The entity already exists, as it's read from SQL properties.

		foreach ($this->_getDefinition()->getProperties() as $propertyName => $propertyDefinition)
		{ // For each defined property, try to read its data.
			// Prepend property name by the prefix.
			$propertyName = $prefix.$propertyName;

			// Try to read only if a value is present.
			if (isset($sqlProperties->$propertyName))
			{ // Get property type.
				if (!($propertyDefinition instanceof FieldBlueprint || $propertyDefinition instanceof ArrayBlueprint))
					throw new UnhandledPropertyTypeException($propertyName, get_class($propertyDefinition));

				// Get property type instance.
				$propertyType = $propertyDefinition->getTypeInstance();

				if (!$propertyType instanceof SqlType)
					// Type is not an SQL type, throwing an exception.
					throw new IncompatibleTypeException(get_class($propertyType), SqlType::class);

				if ($propertyDefinition instanceof FieldBlueprint)
				{ // Read direct value.
					// Parsing property value and put it in the entity.
					$value = $propertyType->fromSql($sqlProperties->$propertyName);
				}
				else
				{ // Read JSON array value.
					$value = json_decode($sqlProperties->$propertyName);
					// Parsing array values.
					foreach ($value as &$val) $val = $propertyType->fromSql($val);
				}

				// Set original property value.
				$this->setOriginalProperty($propertyName, $value);
			}
		}
	}

	/**
	 * Export properties to a raw SQL object.
	 * @return array<string, Raw|string|int|float|null> Raw SQL array where properties are extracted.
	 * @param string $prefix Optional prefix of properties in SQL.
	 * @throws IncompatibleTypeException
	 */
	public function toSqlProperties(string $prefix = ""): array
	{
		// Initialize raw SQL array.
		$object = [];

		foreach ($this->_getDefinition()->getProperties() as $propertyName => $propertyDefinition)
		{ // Get SQL value of each defined property.
			// Prepend property name by the prefix.
			$propertyName = $prefix.$propertyName;

			if (!($propertyDefinition instanceof FieldBlueprint))
				// Ignore all properties that are not normal fields.
				continue;

			if (isset($this->$propertyName))
			{ // If property is set, converting it.
				// Get property type.
				$propertyType = $propertyDefinition->getTypeInstance();

				if (!$propertyType instanceof SqlType)
					// Type is not an SQL type, throwing an exception.
					throw new IncompatibleTypeException(get_class($propertyType), SqlType::class);

				// Converting property value and put it in the SQL object.
				$object[$propertyName] = $propertyType->toSql($this->$propertyName);
			}
		}

		return $object; // Return built raw SQL array.
	}

	/**
	 * Generate columns SELECT for the defined properties.
	 * @param string $prefix Prefix to add to the name of the properties.
	 * @return Raw[] Selected columns.
	 * @throws UnknownDatabaseException
	 * @throws MissingConditionValueException
	 */
	public function sqlSelectFields(string $prefix = ""): array
	{
		$select = [];

		foreach ($this->_getDefinition()->getProperties() as $propertyName => $property)
		{ // For each field, generate a column selection.
			if ($property instanceof FieldBlueprint)
			{ // It's a field property, generate a SELECT for it.
				$select[] = new Raw("\"{$this->getTableName()}\".\"$propertyName\"".(!empty($prefix) ? " AS \"$prefix$propertyName\"" : ""));
			}
			elseif ($property instanceof ArrayBlueprint)
			{ // It's an array field, generate a SELECT for it.
				$select[] = (new ArrayInstance($property))->genSelect($this->query(), $prefix);
			}
		}

		return $select;
	}
}
